#  PommeKit - Python library with various tools for interacting with Apple services and APIs
#  Copyright (C) 2024  Cypheriel

"""Apple Push Notification service (APNs) credential provisioning service — "Albert"."""
