#  PommeKit - Python library with various tools for interacting with Apple services and APIs
#  Copyright (C) 2024  Cypheriel

"""General utilities designed to facilitate connections with the Apple Push Notification service (APNs)."""
