#  PommeKit - Python library with various tools for interacting with Apple services and APIs
#  Copyright (C) 2024  Cypheriel
"""Package for device-related classes and enums."""

from ._device import Device, OperatingSystem

__all__ = [
    "Device",
    "OperatingSystem",
]
